package Stanford::SecureAfsRequest ;

## no critic (Subroutines::RequireArgUnpacking) ;
## no critic (CodeLayout::ProhibitParensWithBuiltins) ;
## no critic (ErrorHandling::RequireCheckingReturnValueOfEval);

use strict ;
use warnings ;
use autodie ;

use Carp;
use CGI qw(:standard);
use Date::Manip qw(ParseDate DateCalc Date_Cmp);
use English;
use File::Spec;
use List::MoreUtils qw(any uniq);
use Net::Remctl;
use Net::SMTP;
use POSIX qw(strftime);
use Text::Template;
use YAML::Tiny;

# This module is needed when sending authenticated SMTP requests.
use Authen::SASL qw(Perl);

use Stanford::Directory;
use Stanford::Infrared::General qw(get_db_setup);
use Stanford::Orange::Util qw(kerberos_ticket_cache kerberos_ticket_cache_no_afs);

# Make this a Moose class
use Moose;
use namespace::autoclean;

# $TEMPLATE_DIR
has template_dir => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/usr/share/secure-afs-request/templates/',
    );

has page => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    );

has template => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has cgi => (
    is       => 'rw',
    isa      => 'Any',
    required => 0,
    );

# The "From:" address on all e-mails.
# $FROM
has from_address => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'afs-request@stanford.edu',
    );


# Set to 1 when running in testing mode. This will ensure
# that the database is not actually updated and that e-mails are sent
# to a test account.
# $TESTMODE
has testmode => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 1,
    );


# When in testmode (see above) all e-mails will be sent to this
# address.
# $TEST_EMAIL
has test_address => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'adamhl@stanford.edu',
    );

# The path to the Kerberos keytab file.
# $KEYTAB
has keytab => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/etc/secure-afs-request/keytab',
    );

# $VERBOSE
has verbose => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
    );

# URL for the group management page
# $MANAGEURL = ;
has manageurl => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/manage-secure-group',
    );


has config_file => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/etc/secure-afs-request/config.yaml',
    );


# Database properties

has db_config_file => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/etc/secure-afs-request/database.ini',
    );

has dsn => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has db_user => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has db_password => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has dbh => (
    is       => 'rw',
    isa      => 'Any',
    required => 0,
    );

# Max number of members per workgroup before warning. We move the exception
# list from AFS into the local file system. Thus, some external process
# will need to manage this file.
# Original location: /afs/ir/service/afs/data/secureafs-wg-exceptions

# $WG_MAX_MEMBERS
has wg_max_members => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    default  => '25',
    );

# $WGEXCEPTIONS
has wgexceptions => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    default  => '/etc/secure-afs-request/secureafs-wg-exceptions',
    );

has remctl_server_afs_info => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has remctl_server_afs_manage => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

has remctl_server_wg_int => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    );

# SMTP properties

has smtp_server => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'smtp.stanford.edu',
    );

has smtp_auth_sends => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
    );

has smtp_auth_credentials_file => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => '/etc/secure-afs-request/smtp-credentials.yaml',
    );

has smtp_auth_principal => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'service/something',
    );

has smtp_auth_password => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
    default  => 'secretsecret',
    );

# Set this to 1 if the host running this application
# is a Stanford AFS client in which case we can run the "vos" directly
# rather than making a remctl call.
has is_afs_client => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
    );


sub BUILD {
    my $self = shift;

    my $template_base = $self->page() . '.tmpl';
    my $template = File::Spec->catfile(($self->template_dir()), $template_base);

    my $cgi = CGI->new;
    $self->cgi($cgi) ;

    $self->template($template);

    # Determine if we are running on a host that is a Stanford AFS client.
    # We do this by seeing if the "vos" command exists.
    my $vos_bin = '/usr/bin/vos';
    if (-f $vos_bin) {
	$self->is_afs_client(1);
    } else {
	$self->is_afs_client(0);
    }

    return;
}

###########################################################################

# Print out error messages in place of web form.
sub print_error {
    my $self = shift;

    my ($error) = @_;
    my $FORM = CGI::FormBuilder->new (
        template => $self->template(),
        submit   => 'Error',
        header   => 1
        );
    $FORM->tmpl_param('badrequest-error' => $error);
    $FORM->tmpl_param('badrequest' => 1);
    print $FORM->render;
    exit;
}

sub gssapi_init {
    my $self = shift;

    $self->progress('about to perform a GSSAPI initialization');

    if (!$self->keytab) {
        croak 'missing keytab definition' ;
    } elsif (! -f $self->keytab) {
        croak 'no keytab file at ' . $self->keytab ;
    }

    my $krb5_cc_file = Stanford::Orange::Util::tempfile() ;

    my $result;
    if ($self->is_afs_client()) {
        $self->progress('we are an AFS client so get AFS token');
        $result = kerberos_ticket_cache($self->keytab, $krb5_cc_file) ;
    } else {
        $self->progress('we are NOT an AFS client so do NOT get AFS token');
        $result = kerberos_ticket_cache_no_afs($self->keytab, $krb5_cc_file) ;
    }

    $self->progress('finished GSSAPI initialization');

    return $result ;
}

sub at_least_one_match {
    my $self = shift;
    my ($regex, $array_ref) = @_ ;

    return any { m/$regex/xsm } @{$array_ref} ;
}

# Return all workgroups for which the user is an administrator.
sub find_administered_workgroups {
    my $self = shift;

    my ($person) = @_;

    if (!$person) {
        croak 'person is not defined (problem with REMOTE_USER?)';
    }

    $self->progress("about to find administered workgroups for user $person");

    # Set up the LDAP connection.
    my $dir = Stanford::Directory->new;
    $dir->set(ldap_server  => 'ldap.stanford.edu',
              mechanism    => 'GSSAPI',
              basedn       => 'cn=people,dc=stanford,dc=edu',
              net_ldapapi  => 1,
        );

    my @privs;
    my @entries = $dir->ldap_query("(uid=$person)");
    foreach my $entry (@entries) {
        foreach my $attr (keys (%{$entry})) {
            if ($attr eq 'suprivilegegroup') {
                foreach (@{$entry->{$attr}}) {
                    if (m{^(.+)/administrator$}xsm) {
                        push(@privs, $1);
                    }
                }
            }
        }
    }

    # Alphabetize.
    @privs = sort @privs;

    my $num_groups = scalar(@privs);
    $self->progress("found $num_groups administered workgroups for user $person");

    return @privs;
}

# Given a workgroup, return all administrators.
sub find_workgroup_admins {
    my $self = shift;
    my ($wg) = @_;

    # Set up the LDAP connection.
    my $dir = Stanford::Directory->new;
    $dir->set (ldap_server  => 'ldap.stanford.edu',
               mechanism    => 'GSSAPI',
               basedn       => 'cn=people,dc=stanford, dc=edu',
               net_ldapapi  => 1,
        );

    my @entries = $dir->ldap_query ("(suPrivilegeGroup=$wg/administrator)", 'uid');
    my @admins = ();
    foreach (@entries) {
        push @admins, $_->uid;
    }

    $self->progress('workgroup admins are ' . join(q{,}, @admins));

    return @admins;
}

sub sql_datetime {
    my $self = shift;

    my ($now) = @_;

    if (!defined($now) || (length($now)==0)) {$now = time;}

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($now);
    $mon++;
    $mon  = sprintf('%02d',$mon);
    $mday = sprintf('%02d',$mday);
    $hour = sprintf('%02d',$hour);
    $min  = sprintf('%02d',$min);
    $sec  = sprintf('%02d',$sec);
    $year += 1900;

    return "$year-$mon-$mday $hour:$min:$sec";
}

# Generic mailer routine that tries to find sendmail and send the message.
sub mail_old {
    my $self = shift;
    my ($mail) = @_ ;

    my $from = $self->from_address() ;

    my ($sendmail) = grep { -x } qw(/usr/sbin/sendmail /usr/lib/sendmail);
    $sendmail ||= '/usr/lib/sendmail';

    open (my $MAIL, "| $sendmail -f $from -t -oi -oem");
    print {$MAIL} $mail;
    close($MAIL);

    return;
}

sub mail {
    my $self = shift;
    my ($mail, $to) = @_ ;

    if (!$to) {
        ## no critic (ValuesAndExpressions::RequireInterpolationOfMetachars);
	croak 'missing required $to parameter';
    }

    my $smtp_server = $self->smtp_server;

    $self->progress("creating Net::SMTP mailer object: $smtp_server");
    my $mailer = Net::SMTP->new($smtp_server, 'SSL' => 1, 'Port' => 465, 'Debug' => 1);
    if (!$mailer) { croak $@ ; }

    # Are we doing authenticated SMTP sends?
    my $smtp_auth_sends = 0;
    if (defined($self->smtp_auth_sends)) {
	$smtp_auth_sends = $self->smtp_auth_sends;
	$self->progress('we are doing an authenticated SMTP send');
    } else {
	$self->progress('we are NOT doing an authenticated SMTP send');
    }

    if ($smtp_auth_sends) {
	my $principal = $self->smtp_auth_principal;
	my $password  = $self->smtp_auth_password;

	$self->progress("authenticating to SMTP server $smtp_server with principal '$principal'");

        # Since we are doing an authenticated send we need to disable any
        # GSSAPI context. This is so the SASL code does not attempt to use
        # any current GSSAPI context rather than using our username and
        # password.
        local $ENV{'KRB5CCNAME'} = '/dev/null';

	if (!$mailer->auth($principal, $password)) {
	    my $status  = $mailer->status();
	    my $message = $mailer->message();
	    carp "error authenticating; status is '$status'";
	    carp "error authenticating; message is '$message'";
	    $self->print_error('Giving up as SMTP authentication failed; please inform support.');
	} else {
	    $self->progress('SMTP authentication successful');
	}
    }

    my $from = $self->from_address() ;

    $self->progress("about to send e-mail to '$to'");

    $mailer->mail($from) ;
    $mailer->to($to) ;
    $mailer->data();
    $mailer->datasend($mail);
    $mailer->dataend();
    $mailer->quit;

    $self->progress('finished e-mail send');

    return;
}


# Issue a command to a given server using remctl and make sure that it is
# successful.  Returns the status of remctl in a scalar context and a list
# containing the status and any output in a list context.
sub make_remctl {
    my $self = shift ;

    my ($server, @command) = @_;

    if ($self->testmode()) {
        # In test mode we don't actually call the remctl, but we do return success
        # so the caller thins the remctl worked.
        return wantarray ? (0, q{}) : 0;
    }

    my ($stdout, $stderr, $status);
    my $result = remctl($server, undef, undef, @command);

    if ($result->error) {
        $self->print_error('remctl failed with error ', $result->error);
        return 1;
    } else {
        $stderr = $result->stderr;
        $stdout = $result->stdout;
        $status = $result->status;
    }

    return wantarray ? ($status, $stdout) : $status;
}

# Check if given workgroup is already in metadata db. If yes, return
# associated group name.
sub wg_already_in_db {
    my $self = shift ;

    my ($wg) = @_;

    my $volname = q{};

    $self->db_connect();
    my $query = 'SELECT volname FROM volumes WHERE workgroup=? ';

    eval {
        my $sth = $self->dbh()->prepare("$query");
        $sth->execute($wg);
        ($volname) = $sth->fetchrow_array;
        $sth->finish;
    };
    $self->db_disconnect();

    if ($volname) {
        $self->progress("found volume $volname corresponding to workgroup $wg");
    } else {
        $self->progress("no volume corresponding to workgroup $wg found");
    }

    if (defined($volname) && ($volname =~ /^secure[.](.+)/xsm)) {
        # Return group name only instead of full volume name.
        return $1;
    }
    else {
        return 0;
    }
}

# Check if given volume name is already in metadata db. If yes, return
# associated workgroup. If not, return the number zero.
sub volname_already_in_db {
    my $self = shift ;

    my ($volname) = @_;
    $self->progress("volname is $volname");

    my $wg = q{};

    $self->db_connect();
    my $query = 'SELECT workgroup FROM volumes WHERE volname=? ';

    $self->progress("$query");

    eval {
        my $sth = $self->dbh()->prepare("$query");
        $sth->execute($volname);
        ($wg) = $sth->fetchrow_array;
        $sth->finish;
    };
    my $at_error = $EVAL_ERROR;

    if ($wg) {
        $self->progress("found workgroup $wg corresponding to volname $volname");
    } else {
        $self->progress("no workgroup corresponding to volname $volname found");
    }

    $self->db_disconnect();

    if (defined($wg) && ($wg =~ /\w/xsm)) {
        return $wg;
    } else {
        return 0;
    }
}

# Takes a workgroup as an argument and determines if there are more than
# the maximum recommended number of members.
sub wg_exceeds_max {
    my $self = shift ;
    my ($wg) = @_;

    my (@exceptions) = ();

    # Set up the LDAP connection.
    my $dir = Stanford::Directory->new;
    $dir->set (ldap_server  => 'ldap.stanford.edu',
               mechanism    => 'GSSAPI',
               basedn       => 'cn=people,dc=stanford, dc=edu');

    my @members = $dir->ldap_query ("(suPrivilegeGroup=$wg)", 'uid');

    my $num = scalar @members;

    # Check if wg is in exceptions list. It's exempt from the max
    # membership limit if it is.
    open(my $IN, '<', $self->wgexceptions());
    @exceptions = <$IN>;
    close ($IN);
    if (at_least_one_match(qr/^$wg$/xsm, \@exceptions)) {
        return 0;
    }
    if ($num >= $self->wg_max_members()) {
        return $num;
    }
    else {
        return 0;
    }
}

# Use SUNet ID to look up user's workgroups and then check all workgroups
# against secure afs db for an associated AFS group.
sub get_groups {
    my ($self) = shift ;

    my ($groups, $query) = (q{});
    my @workgroups = ();

    my $USER = $ENV{'REMOTE_USER'};

    # Get user's workgroup membership.
    @workgroups = $self->find_administered_workgroups($USER);

    # Connect to database.
    $self->db_connect();

    # For each workgroup, provide link to associated secure afs group if
    # any.
    my @vols = ();
    foreach my $wg (@workgroups) {
        chomp $wg;

        # Get volume name associated with workgroup.
        $query = 'SELECT volname FROM volumes '
            . 'WHERE workgroup=? ';
        my $sth = $self->dbh()->prepare($query);
        eval {
            $sth->execute($wg);
            while ( my $row = $sth->fetchrow_hashref ) {
                my $volname = $row->{volname};
                $volname =~ s/^secure[.]//xsm;
                push @vols, $volname;
            }
        };
    }
    $self->db_disconnect();

    foreach my $vol (sort @vols) {
        my ($wg, $path, $grace, $exp_dt, $contact, $quota) = $self->get_prefs($vol);

        # Strip off time to isolate date
        $exp_dt =~ s/\s.*//gxsm;

        my $manageurl = $self->manageurl();

        my $title= qq{title="Manage volume $vol"};

        if ($self->expired($exp_dt)) {
            $groups .= "<li><a href=\"$manageurl?group=$vol\" $title>$vol</a>"
                . " <font size='-2' color='#990000'><b>(EXPIRED)</b></font><br></li>\n";
        }
        else {
            $groups .= "<li><a href=\"$manageurl?group=$vol\" $title>$vol</a><br></li>\n";
        }
    }
    if (! length($groups) ) {
        $groups = '<li><i>You currently do not have access to edit settings '
            . "for a secure AFS group.</i></li>\n";
    }

    return "<ul>$groups</ul>";
}


# Given a group name, return associated workgroup, file path, grace
# period, exp date, and contact address.
sub get_prefs {
    my ($self) = shift ;

    my ($group) = @_;
    my ($wg, $path, $grace, $exp_dt, $contact) = (q{});
    my $quota = 0;

    $self->db_connect();
    my $query = 'SELECT workgroup, path, grace_period, exp_dt, '
        . 'contact_email, quota FROM volumes WHERE volname=? ';
    my $volname = 'secure.' .$group;
    eval {
        my $sth = $self->dbh()->prepare("$query");
        $sth->execute($volname);
        ($wg, $path, $grace, $exp_dt, $contact, $quota) = $sth->fetchrow_array;
        $sth->finish;
    };
    $self->db_disconnect();

    # Display quota in GBs.
    ## no critic (ValuesAndExpressions::ProhibitMagicNumbers);
    if ($quota > 0) {
        $quota = $quota / 1000;
    }

    return ($wg, $path, $grace, $exp_dt, $contact, $quota);
}


# Determines if group is already expired and in its grace period.
sub expired {
    my ($self) = shift ;
    my ($exp_dt) = @_;

    my $now  = strftime ('%m/%d/%Y', localtime(time));
    my $d1   = ParseDate($exp_dt);
    my $d2   = ParseDate($now);
    my $flag = Date_Cmp($d2,$d1);

    # If $flag is less than 0, then now is earlier than
    # exp date and the group has not expired.
    return ($flag < 0) ? 0 : 1;
}

# Takes the expiration date as an argument and determines if account
# will expire within approximately one month.
sub expires_soon {
    my $self = shift ;
    my ($exp_dt) = @_;

    my $now = strftime ('%m/%d/%Y', localtime(time));

    my $d1   = ParseDate($exp_dt);
    my $d2   = ParseDate($now);
    my $new  = DateCalc($d2, '+ 31 days');
    my $flag = Date_Cmp($new, $d1);

    # If $flag is less than 0, then now + 31 days is still earlier than
    # exp date.
    return ($flag < 0) ? 0 : 1;
}

sub db_connect {
    my $self = shift ;

    my ($dsn, $db_user, $db_pass) = get_db_setup($self->db_config_file());

    if (!$self->dbh()) {
        my $dbh = DBI->connect ($dsn, $db_user, $db_pass,
                                {RaiseError => 0,
                                 AutoCommit => 1});
        ## no critic (Variables::ProhibitPackageVars);
        if ($DBI::err) {
            croak "Connecting to database failed: $DBI::errstr";
        } else {
            $self->dbh($dbh) ;
        }
    }

    return 1;
}

sub db_disconnect {
    my $self = shift ;

    eval {$self->dbh()->disconnect()};
    $self->dbh(undef);
    return;
}

sub progress {
    my $self = shift ;
    my ($msg) = @_ ;

    if ($self->verbose()) {
        carp "progress: $msg" ;
    }
    return ;
}

sub template_path {
    my $self = shift ;
    my ($basename) = @_ ;
    return File::Spec->catfile(($self->template_dir()), $basename);
}

sub read_config_file {
    my $self = shift ;

    my ($config_file) = @_ ;

    if (!$config_file) {
	$config_file = $self->config_file() ;
    }

    my $yaml = YAML::Tiny->read($config_file);

    my $config = $yaml->[0];

    # All possible configuration file directives (some may not be
    # present).
    my @attributes = qw(
	verbose
	testmode
	test_address
	db_config_file
	from_address
	keytab
	remctl_server_afs_info
	remctl_server_afs_manage
	remctl_server_wg_int
        smtp_server
	smtp_auth_sends
	smtp_auth_credentials_file
    ) ;

    foreach my $attribute (@attributes) {
	if (exists($config->{$attribute})) {
	    my $value = $config->{$attribute};
	    $self->progress("setting attribute '$attribute' to value '$value' from config file");
	    $self->$attribute($value);
	}
    }

    # Some of these are required, so fail if they are not defined.
    my @attributes_required = qw(
        db_config_file
        from_address
        keytab
        remctl_server_afs_info
        remctl_server_afs_manage
        remctl_server_wg_int
        ) ;

    foreach my $attribute (@attributes_required) {
        if (!$self->$attribute) {
            croak "required attribute '$attribute' not defined";
        }
    }

    # If we are doing authenticated SMTP sends, read the file containing the
    # username and password.
    if ($self->smtp_auth_sends) {
	$self->read_smtp_credentials();
    }

    return ;
}

sub read_smtp_credentials {
    my $self = shift ;

    my $smtp_auth_credentials_file = $self->smtp_auth_credentials_file;
    $self->progress("reading smtp authentication credentials from '$smtp_auth_credentials_file'");

    my $yaml   = YAML::Tiny->read($smtp_auth_credentials_file);
    my $config = $yaml->[0];

    my $principal = $config->{'smtp_auth_principal'};
    my $password  = $config->{'smtp_auth_password'};

    $self->progress("setting attribute 'smtp_auth_principal' to value '$principal' from file '$smtp_auth_credentials_file'");
    $self->smtp_auth_principal($principal);
    $self->progress("setting attribute 'smtp_auth_password' to value 'REDACTED' from file '$smtp_auth_credentials_file'");
    $self->smtp_auth_password($password);

    return;
}


__PACKAGE__->meta->make_immutable;

1 ;
